#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from test_elasticsearch.test_cases import ElasticsearchTestCase


class TestIndices(ElasticsearchTestCase):
    def test_create_document(self):
        self.client.create(index="test-index", id="test-id", document={"key": "value"})
        self.assert_url_called("PUT", "/test-index/_create/test-id")

    def test_create_document_with_type(self):
        self.client.create(
            index="test-index",
            id="test-id",
            doc_type="test-type",
            document={"key": "value"},
        )
        self.assert_url_called("PUT", "/test-index/test-type/test-id/_create")
